from __future__ import annotations

from hats.catalog.healpix_dataset.healpix_dataset import HealpixDataset


class MapCatalog(HealpixDataset):
    """A HATS table to represent non-point-source data in a continuous map."""
